
type 'a tas = {
  donnees : 'a array;
  mutable taille : int
}

(* Renvoie l'indice du fils gauche de l'élément d'indice i *)
let gauche i = 2 * i + 1

(* Renvoie l'indice du fils droit de l'élément d'indice i *)
let droite i = 2 * i + 2

(* Renvoie l'indice du parent de l'élément d'indice i *)
let parent i = (i - 1) / 2


let tas_creer capacite elem = {
  donnees = Array.make capacite elem; 
  taille = 1;
}

let tas_echanger (tas : 'a tas) (i : int) (j : int) =
  assert (0 <= i && i < tas.taille);
  assert (0 <= j && j < tas.taille);
  let tmp = tas.donnees.(i) in
  tas.donnees.(i) <- tas.donnees.(j);
  tas.donnees.(j) <- tmp


let rec tas_percole_haut tas i =
  if i > 0 then
    let p = parent i in
    if tas.donnees.(i) < tas.donnees.(p) then
      tas_echanger tas i p;
      tas_percole_haut tas p
  
let rec tas_percole_bas tas i =
  let g = gauche i in
  let d = droite i in
  let n = tas.taille in

  let min_index =
    if g < n && tas.donnees.(g) < tas.donnees.(i) then g else i
  in
  let min_index =
    if d < n && tas.donnees.(d) < tas.donnees.(min_index) then d else min_index
  in

  if min_index <> i then begin
    tas_echanger tas i min_index;
    tas_percole_bas tas min_index
  end
      

let tas_inserer (tas: 'a tas) (element: 'a) =
  if tas.taille < Array.length tas.donnees then begin
    tas.donnees.(tas.taille) <- element;
    tas.taille <- tas.taille + 1;
    tas_percole_haut tas (tas.taille - 1)
  end else
    failwith "Tas plein"



let tas_extraire (tas: 'a tas) : 'a =
  if tas.taille = 0 then
    failwith "Tas vide"
  else begin
    let min_elem = tas.donnees.(0) in
    tas.donnees.(0) <- tas.donnees.(tas.taille - 1);
    tas.taille <- tas.taille - 1;
    tas_percole_bas tas 0;
    min_elem
  end
    



let dijkstra graphe s t =
  (* Initialisation des structures dont on a besoin *)
  let n = Array.length graphe in
  let dist = Array.make n infinity in
  let vu = Array.make n false in
  let a_voir = tas_creer (n*n) (0.0, s) in (* a_voir est le tas binaire *)
  dist.(s) <- 0.0;

  (* Boucle principale *)
  while a_voir.taille > 0 do
    let (dv, v) = tas_extraire a_voir in
      if not vu.(v) then begin
        vu.(v) <- true;
        for w=0 to n-1 do
          if v <> w && graphe.(v).(w) >= 0. then
            let d = dv +. graphe.(v).(w) in
            if d < dist.(w) then begin 
              dist.(w) <- d;
              tas_inserer a_voir (d, w)
            end
        done
      end;
      for i=0 to a_voir.taille-1 do
        Printf.printf "prio=%f, s=%d;\t" (fst a_voir.donnees.(i)) (snd a_voir.donnees.(i))
      done;
      Printf.printf "\n"
  done;

  (* Conclusion *)
  if dist.(t) = infinity then failwith "On ne peut pas atteindre la cible"
  else dist.(t)


let graphe1 = [|
  [|-1.;85.;217.;-1.;173.;-1.;-1.;-1.;-1.;-1.|];
  [|85.;-1.;-1.;-1.;-1.;80.;-1.;-1.;-1.;-1.|];
  [|217.;-1.;-1.;-1.;-1.;-1.;186.;103.;-1.;-1.|]; 
  [|-1.;-1.;-1.;-1.;-1.;-1.;-1.;183.;-1.;-1.|];
  [|173.;-1.;-1.;-1.;-1.;-1.;-1.;-1.;-1.;502.|]; 
  [|-1.;80.;-1.;-1.;-1.;-1.;-1.;-1.;250.;-1.|];
  [|-1.;-1.;186.;-1.;-1.;-1.;-1.;-1.;-1.;-1.|]; 
  [|-1.;-1.;103.;183.;-1.;-1.;-1.;-1.;-1.;167.|];
  [|-1.;-1.;-1.;-1.;-1.;250.;-1.;-1.;-1.;84.|]; 
  [|-1.;-1.;-1.;-1.;502.;-1.;-1.;167.;84.;-1.|]
|]

let d1 = dijkstra graphe1 0 9


let graphe2 = [|
  [|-1.;1.;1.;-1.;-1.;-1.|];
  [|1.;-1.;3.;-1.;-1.;10.|]; 
  [|1.;3.;-1.;2.;4.;-1.|];
  [|-1.;-1.;2.;-1.;-1.;4.|]; 
  [|-1.;-1.;4.;-1.;-1.;3.|]; 
  [|-1.;10.;-1.;4.;3.;-1.|]
|]

let d2 = dijkstra graphe2 0 5


let graphe3 = [|
  [|-1.;1.;3.;5.;7.|];
  [|1.;-1.;1.;3.;5.|]; 
  [|3.;1.;-1.;1.;3.|];
  [|5.;3.;1.;-1.;1.|];
  [|7.;5.;3.;1.;-1.|]
|]

let d3 = dijkstra graphe3 0 4